from dnalib.utils import TableUtils, Utils
from .tablecomment import TableComment
from dnalib.log import log

class CreateView:
    """
        Classe que implementa a criação de views voltadas ao consumo da área de negócios. As views fazem hash dos campos indicados como anonimizados.

        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            layer (str): camada da tabela no lake.   
            table_name (str): string que representa o nome da tabela.
            yml (yml object or str, optional): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create View. Padrão é None.
            parsed_comment (str, optional): string com comentário parseado a partir da TableComment. Se não for informado é obtido a partir de uma istância de TableComment.
            anonimized_fields (list, optional): lista de campos para serem anonimizados.            
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            layer (str): camada da tabela no lake.   
            table_name (str): string que representa o nome da tabela.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            parsed_comment (str): string com comentário parseado.
            anonimized_fields (list): lista de campos para serem anonimizados.
            fields (dict): dicionário com o padrão field_name:comment. 
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            parsed_template (str): template do comentário da tabela compilado, preenchido após a chamada do método parse.
            parsed_fields (str): string com os comentários das fields parseados no padrão do Create View, preenchido após a chamada do método parse_fields().
    """

    def __init__(self,                  
                 layer, 
                 table_name,                  
                 yml={},                                    
                 tbl_comment=None,
                 anonimized_fields=[],   
                 comment={},
                 comment_squad={},               
                 fields={},
                 replace=False):                                 
        self.yml = yml
        self.replace = replace 
        # yml must be either a string or a dict from the yml file        
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.table_name = table_name.strip().lower()        
        self.layer = layer.strip().lower()           
        self.tbl_comment = tbl_comment        
        self.parsed_comment = ""
        if self.tbl_comment == None: 
            self.tbl_comment = TableComment(layer, table_name, yml, anonimized_fields, comment, comment_squad)
        if self.tbl_comment.parsed_comment == "":
            self.tbl_comment = self.tbl_comment.parse()        
        self.parsed_comment = self.tbl_comment.parsed_comment
        self.anonimized_fields = self.yml.get("anonimized_fields", anonimized_fields)        
        self.fields = self.yml.get("fields", fields)
        self.field_comments = {}
        self.parsed_template = ""
        self.parsed_fields = ""
    
    # by default the python parameter overwrite the yml if both are informed
    def __load_and_verify_params(self):
        """
            Método interno que carrega as estruturas principais para popular a estrutura da view.

            Raises:
                ValueError: Ocorre se a estrutura fields não for passada. Ou deve se passada por parâmetro, ou deve estar dentro do yml.
        """             
        if len(self.fields) == 0:
            self.fields = TableUtils.table_fields_metadata(self.layer, self.table_name)                        
            if len(self.fields) == 0:
                log(__name__).error(f"If not informed or infered, the fields parameter must be in yml table template for {self.layer} layer.")
                raise ValueError(f"If not informed or infered, the fields parameter must be in yml table template for {self.layer} layer.")  
        
        if len(self.anonimized_fields) == 0:
            self.anonimized_fields = self.tbl_comment.anonimized_fields

        # manually add dataCarga and checksum comments (default values)
        #self.field_comments["dataCarga"] = ["Data de carga dos dados."]
        self.fields["dataCarga"] = ["Data de carga dos dados."]
        self.field_comments["checksum"] = ["CheckSum/Hash das informações do registro (linha)."]

        # populate comments structure
        for field_name, field_params in self.fields.items():                                    
            self.field_comments[field_name] = field_params[0]

    def parse(self):
        """
            Método que faz o parse dos comentários baseados na estrutura do Atlan. 

            Returns:
                self: a instância da classe CreateView.
        """

        self.__load_and_verify_params()
        list_of_parsed_fields = ["{} COMMENT '{}'".format(field_name, TableUtils.format_comment_content(self.field_comments[field_name])) for field_name in self.fields]
        self.parsed_fields = ", ".join(list_of_parsed_fields)
        return self        
    
    def template(self):   
        """
            Método que constroí o template da estrutura da view.

            .. code-block:: SQL

                CREATE OR REPLACE VIEW layer.table_name_vw (parsed_fields)                
                    COMMENT "parsed_comment"
                AS
                    SELECT fields FROM layer.table_name
            
            Returns:
                self: a instância da classe CreateView.
        """

        # parsed fields
        self.parse()   
        
        list_of_fields = []
        # take each field and format do hash it
        for field in self.fields:            
            list_of_fields.append(TableUtils.hash_field_str(field) if field in self.anonimized_fields else field)
          
        self.parsed_template = """
            CREATE OR REPLACE VIEW {}.{}_vw             
            ({}) 
            COMMENT "{}"
            AS
            SELECT {} FROM {}.{}
        """.format(self.layer, self.table_name, self.parsed_fields, self.parsed_comment, ", ".join(list_of_fields), self.layer, self.table_name)         
        return self
    
    def execute(self):        
        """
            Método que executa todo processo para gerar e escrever a view.              
            
            Returns:                
                self: a instância da classe CreateView.

            Raises:
                ValueError: Se a tabela informada não existir.
        """
        
        # verify if table exists
        if not TableUtils.table_exists(self.layer, self.table_name):
            log(__name__).error(f"Table {self.layer}.{self.table_name} does not exist")
            raise ValueError(f"Table {self.layer}.{self.table_name} does not exist")

        if not TableUtils.view_exists(self.layer, f"{self.table_name}_vw") or self.replace:
            # generate final template
            self.template()

            # creating view
            Utils.spark_instance().sql(self.parsed_template)  
        else:
            log(__name__).warning(f"The view {self.layer}.{self.table_name}_vw already exists, so nothing will be done.")
        return self



from pyspark.sql.functions import col, lit
from dnalib.utils import TableUtils, Utils
from dnalib.log import log
from dnalib.tables import DynamicTable
from dnalib.writer import BatchTableWriter
from dnalib.core import Table
from .block import Block
from delta.tables import DeltaTable

class BlockChain(Table,
                 BatchTableWriter,
                 DynamicTable):
    """
        Implementação da blockchain
    """

    layer = "blockchain"

    def __init__(self, table_name):            
        # chamando o super para criar a tabela
        Table.__init__(self, self.layer, table_name, Block.schema)          
        BatchTableWriter.__init__(self, self.layer, table_name)        
        self.create()        
        # carragndo dados da tabela de blockchain
        self.df_block_chain = self.spark.table(self.table_catalog_path) 
        # armazenando a computação corrente da blockchain
        self.df_block_chain_current = None 
        # chamando o super para a DynamicTable
        DynamicTable.__init__(self, self.layer, table_name, json_fields=["payload"])                       

    def compute_blockchain(self, df_block, block_name, prev_block_name=None, prev_block_join_rules=None):     
        # cleaning the current blockchain to allow new computations
        self.df_block_chain_current = self.spark.createDataFrame([], Block.schema)  
        # usa a blockchain como o "bloco anterior"
        df_block_chain = self.df_block_chain.alias("block_prev")      
        # colunas que são usadas nas queries  
        columns = df_block.columns
        #
        #   É preciso determinar primeiro quais transações são novas
        #   e isso pode ser feito a partir do id block (hash)
        #   toda transação nova não pode estar na blockchain
        #   
        #   Dado que A é o bloco novo e B a blockchain, 
        #   então A \ B são as novas transações (garantia de unicidade de registro na blockchain).
        #
        df_block = (
            df_block.join(
                df_block_chain, 
                on=[col("block.id_block") == col("block_prev.id_block")], 
                how="leftanti"
        )).alias("block")        
        # blocos que iniciam a cadeia não tem estrutura de prev (são raízes do grafo) 
        if prev_block_join_rules is not None:                 
            # otimizando a consulta filtrando pelo tipo do block anterior (usando o nome), já que todo bloco deve ter apenas um parent
            if prev_block_name is not None:                
                df_block_chain = df_block_chain.filter(col("block_name") == lit(prev_block_name))                
            #
            # Primeira operação é verificar se existe conexão com bloco anterior (usando left anti join)
            # nesse caso os blocos são geradores (ou seja não possuem conexão anterior na cadeia).
            #
            df_next_block = (
                df_block.join(
                    df_block_chain, 
                    on=prev_block_join_rules.field_rep, # chaves de negócio
                    how="leftanti"
            ))                        
            # o bloco é inserido na blockchain       
            df_block_union = df_next_block.select(*columns)     
            self.df_block_chain_current = self.df_block_chain_current.union(df_block_union)
            # agora é verificado os blocks que tem conexão na cadeia (via inner join)
            df_next_block = (
                df_block.join(
                    df_block_chain, 
                    on=prev_block_join_rules.field_rep, # chaves de negócio
                    how="inner"
            ))      
            #
            # As colunas do blockchain são agora propagadas para o próximo bloco:
            #   o id_block_gen vem do bloco anterior
            #   o id_block_prev é o id do bloco anterior 
            #   como conectou, esse bloco não é gerador
            #
            propagate_block_fields = (
                col("block.block_name"),
                (col("block_prev.block_level") + lit(1)).alias("block_level"),
                col("block.block_business_keys"),
                col("block_prev.id_block_gen"),  
                col("block.id_block"),
                col("block_prev.id_block").alias("id_block_prev"),    
                lit(0).alias("is_gen"),
                col("block.block_timestamp"),
                col("block.payload"),
                lit(1).alias("block.last_block")
            )            
            # o bloco é inserido na blockchain                         
            df_block_union = df_next_block.select(*propagate_block_fields)
            self.df_block_chain_current = self.df_block_chain_current.union(df_block_union)                                  
        else:
            # quando o bloco é a "raíz" ele não possui bloco prev, portanto ele é inserido direto na blockchain
            df_block_union = df_block.select(*columns)                   
            self.df_block_chain_current = self.df_block_chain_current.union(df_block_union)           
        return self
    
    def update_last_block(self):              
        # para todo bloco A na cadeia, se existe B tal que B \neq A e id_block_prev(B) = id_block(A), então last_block(A) = 0
        log(__name__).info(f"Updating last_block column to previously blockchain blocks.")
        df_non_last_blocks = self.df_block_chain_current.select(col("id_block_prev").alias("_id_block_prev")).filter("_id_block_prev is not null").distinct()
        # como a computação é incremental, então existem blocos onde o last_block(A) = 0, porém last_block_{-1}(A) = 1 (ou seja, estão salvos na tabela delta como 1)
        df_prev_last_blocks = df_non_last_blocks.alias("block").join(
            self.df_block_chain_current.alias("block_prev"),
            on=[col("block._id_block_prev") == col("block_prev.id_block")],
            how="leftanti",
        )
        # atualizando somente os blocos anteriores
        self.update(df_prev_last_blocks, "source.id_block = update._id_block_prev", set_dict={"last_block":lit(0)})

    # atualiza a tabela dos blocos
    def update_blocks(self):                
         # logando o número de novos blocos
        total_new_blocks = self.df_block_chain_current.count() 
        # atualizando a blockchain
        if total_new_blocks > 0:
            self.update_last_block()                        
            # as novas transações são inseridas na tabela do blockchain
            log(__name__).info(f"Appeding new blocks to blockchain table {self.layer}.{self.table_name}.")
            self.append(self.df_block_chain_current)
        # logging new blocks
        log(__name__).info(f"Total of {total_new_blocks} blocks were added to blockchain.")

    def add_block_bucket(self, block_bucket, list_of_block_names):
        # compute blockchain structure based on a bucket of blocks
        for block_name in list_of_block_names:
            # carregando dados do bucket de acordo com o bloco que está sendo processado
            df_block, prev_block_name, prev_block_join_rules = block_bucket.load(block_name)
            # computando a blockchain
            self.compute_blockchain(df_block.alias("block"), block_name, prev_block_name, prev_block_join_rules)                    
            # atualizar a blockchain
            self.update_blocks()
            # "marcando como processados" os novos blocos da blockchain            
            df_block_in_chain = (df_block.alias("bucket").join(
                self.df_block_chain.alias("blockchain"), 
                on=[col("bucket.id_block") == col("blockchain.id_block")], 
                how="inner")
            )
            block_bucket.update(df_block_in_chain.select(col("bucket.id_block")).distinct(), "source.id_block = update.id_block", set_dict={"is_block_processed":lit(1)})
        # otimizando a tabela ao final das atualizações
        self.optimize()   

    def add_block(self, block, prev_block_name=None, prev_block_join_rules=None):  
        # compute blockchain structure based on single block
        self.compute_blockchain(block.load().df_block.alias("block"), block.block_name, prev_block_name, prev_block_join_rules)      
        # atualizar a blockchain
        self.update_blocks()
             


from pyspark.sql.functions import col, get_json_object
import pickle
import re

class Key:	
    """ 
        A classe Key é um wrapper para o método col do Pyspark, com a adição que ela armazena as fields em listas internas, para consulta.

        Args:
            field (str): string que representa o nome de uma coluna do pyspark, internamente é repassado para o col.

        Attributes:
            field (str): string que representa o nome de uma coluna do pyspark, internamente é repassado para o col.
            field_rep (col): resultado de várias operações booleanas feitas internamente com a função col.
            fields (list): lista com todas as fields resultantes das operações com as Keys.
            lfieds (list): lista com todas as fields a esquerda resultantes das operações com as Keys.
            lfieds (list): lista com todas as fields a direita resultantes das operações com as Keys.

    """ 
    
    def __init__(self, field):
        self.field = field        
        self.field_rep = None
        self.fields = [field]        
        self.lfields = []
        self.rfields = []
        self.key_field_rep = f'Key("{field}")'

    @staticmethod
    def from_serialized(serialized_key):   
        return eval(serialized_key)        
    
    def serialize(self):        
        return self.key_field_rep

    def right_fields(self):
        """
            Método getter que retorna a lista de fields a direita.

            Returns: 
                list: lista com todas as fields a direita.
        """
        return list(dict.fromkeys(self.rfields).keys())

    def left_fields(self):
        """
            Método getter que retorna a lista de fields a esquerda.

            Returns: 
                list: lista com todas as fields a esquerda.
        """
        return list(dict.fromkeys(self.lfields).keys())

    def chain_key(self, df_alias, key_name):
        return get_json_object(col(f"{df_alias}.payload"), f"$.{key_name}")
    
    def propagate_fields(self, other, op):
        # na primeira vez, precisa preencher as fields a esquerda da operação
        if len(self.lfields) == 0:
            self.lfields.append(self.field)

        # na primeira vez, precisa preencher as fields a direita da operação
        if len(self.rfields) == 0:
            self.rfields.append(other.field)        

        # propaga as fields da esquerda
        self.lfields = self.lfields + other.lfields
        # propaga as fields da direita
        self.rfields = self.rfields + other.rfields
        # propaga todos os valores de fields
        self.fields = self.fields + other.fields  

        # criando uma string para representar as operações
        self.key_field_rep = f"({self.key_field_rep} {op} {other.key_field_rep})"        

    def init_fields(self, other):        
        if self.field_rep is None:
            self.field_rep = self.chain_key("block", self.field)        
        if other.field_rep is None:
            other.field_rep = self.chain_key("block_prev", other.field)

    def __repr__(self):
        return str(self.key_field_rep)

    def __str__(self):
        return str(self.key_field_rep)

    def __or__(self, other):                
        self.init_fields(other)
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep | other.field_rep   
        # propaga as fields
        self.propagate_fields(other, "|")       
        return self

    def __and__(self, other):                
        self.init_fields(other)
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep & other.field_rep    
        # propaga as fields
        self.propagate_fields(other, "&")       
        return self
    
    def __lt__(self, other):   
        self.init_fields(other)     
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep < other.field_rep
        # propaga as fields
        self.propagate_fields(other, "<")
        return self
    
    def __le__(self, other):   
        self.init_fields(other)     
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep <= other.field_rep
        # propaga as fields
        self.propagate_fields(other, ">=")
        return self
    
    def __eq__(self, other):   
        self.init_fields(other)     
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep == other.field_rep
        # propaga as fields
        self.propagate_fields(other, "==")
        return self
    
    def __ne__(self, other):   
        self.init_fields(other)     
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep != other.field_rep
        # propaga as fields
        self.propagate_fields(other, "!=")
        return self
    
    def __gt__(self, other):   
        self.init_fields(other)     
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep > other.field_rep
        # propaga as fields
        self.propagate_fields(other, ">")
        return self
    
    def __ge__(self, other):   
        self.init_fields(other)     
        # popula a estrutura de fields usando o col do spark
        self.field_rep = self.field_rep >= other.field_rep
        # propaga as fields
        self.propagate_fields(other, ">=")
        return self


from dnalib.core import CreateTableSilver, TableComment, CreateView
from dnalib.utils import TableUtils, Utils
from dnalib.writer import WriteModes
from dnalib.log import log
from .tables import ProcessLayerTable
from pyspark.sql.functions import col
from enum import Enum

class SilverLoaderType(Enum):
    INCREMENTAL = "incremental" # read data from dataCarga
    FULL = "full" # read all    
    STREAMING = "streaming" # for streaming purpose

class SilverTable(ProcessLayerTable):
    """ 
        A classe SilverTable é uma implementação especializada da classe ProcessLayerTable que implementa todo o pipeline commons que seguimos em uma tabela da silver. Nela é possível carregar o dado da bronze, formatar, fazer cast, criar a tabela e persistir o dado com poucas linhas de código.

        Args:
            table_name (str): string que representa o nome da tabela.
            yml (yml object or str): caminho do arquivo yml ou instância do objeto com os parâmetros necessários.
            source_df (spark Dataframe): dataframe do spark onde os dados serão consumidos para compor o target_df.      
            include_checksum (bool): se verdadeiro adiciona o campo de checksum no target_df.     
            loader_type (SilverLoaderType): tipo de carregamento, se é incremental ou full.

        Attributes:
            layer (str): constante com o valor "silver".
            dict_of_fields (dict): estrutura que armazena as chamadas do método field(), utilizado para garantir que o processo seja lazy (como o spark).
    """

    layer = "silver"

    def __init__(self, source_table_name, yml, target_table_name=None, source_df=None, include_checksum=False, loader_type=SilverLoaderType.INCREMENTAL):
        if target_table_name is None:
            self.target_table_name = source_table_name
        else:
            self.target_table_name = target_table_name
        super().__init__(self.layer, self.target_table_name, source_df, include_checksum)    
        self.source_table_name = source_table_name
        self.yml = yml
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.loader_type = loader_type
        self.source_df_query = self.yml.get("source_df_query", "")
        self.fields = self.yml.get("fields", {})
        self.dict_of_fields = {}     
        self.create_tbl = None
        self.create_vw = None                
        self.__populate_fields()

    def __populate_fields(self):
        # verify if the yml file has the fields parameter
        if len(self.fields) == 0:
            log(__name__).error("The parameter fields is required in the yml file")
            raise ValueError("The parameter fields is required in the yml file")        
        for field_name, field_params in self.fields.items():                        
            self.field(field_name, *field_params[1:]) 
        return self

    def parse_df_source(self):
        """
            Método que executa o carregamento de dados da camada bronze para a silver, o método é baseado no campo de dataCarga. Se a tabela não existir, os dados são carregados de maneira full,
            caso contrário os dados são carregados somente com os registros que foram atualizados.

            Returns:
                source_df (spark DataFrame): dataframe carregado a partir da camada source, caso source_df seja None.

            Raises:
                ValueError: caso a tabela bronze não exista.
        """         
        # verify if table exists in bronze layer
        if not TableUtils.table_exists("bronze", self.source_table_name):
            log(__name__).error(f"Table bronze.{self.source_table_name} does not exists")
            raise ValueError(f"Table bronze.{self.source_table_name} does not exists")        

        # this loads the "center" of the query
        if self.source_df_query == "":
            parsed_fields = ", ".join(self.dict_of_fields.keys())
            self.source_df_query = f"""
                SELECT      
                    {parsed_fields}               
                FROM bronze.{self.source_table_name}
            """
            
        # transform strategy in this case is to check if the table exists and the loader is incremental
        if TableUtils.table_exists(self.layer, self.table_name) and self.loader_type.value == SilverLoaderType.INCREMENTAL.value:
            # verify if it already has a where condition
            has_where = self.source_df_query.upper().find("WHERE")            
            # we will transform based on the last loaded dataCarga
            self.source_df_query += """ {} dataCarga > (SELECT COALESCE(MAX(dataCarga), '1700-01-01') FROM silver.{})
            """.format("AND" if has_where != -1 else "WHERE", self.table_name)           

        return Utils.spark_instance().sql(self.source_df_query)

    def load_fields_from_source(self):
        """
            Método interno que executa o processo de formatar, fazer cast e renomear as fields que serão levadas da bronze para a silver.

            Returns:
                target_fields (list): lista com as fields tratadas que serão consumidas da camada bronze para a silver.
        """        
        source_schema = self.source_df.schema
        target_fields = []
        for source_field_name, target_field_type, target_field_name, format_field in self.dict_of_fields.values():
            field = col(source_field_name)
            # automatic format field 
            formated_field = self.format_field(field, source_schema[source_field_name].dataType.simpleString(), target_field_type, format_field)
            # cast to field type        
            cast_field = self.cast_field(formated_field, target_field_type)
            # final field
            final_field = self.rename_field(cast_field, source_field_name, target_field_name)
            # append final fields
            target_fields.append(final_field)

        # return fields from source        
        return target_fields

    def create_table(self):     
        """
            Esse método é um wrapper para o método CreateTableSilver.execute() e CreateView.execute(). Ele cria a tabela e a view para a camada silver. Deve ser chamado após o load(), caso contrário o último será chamado internamente.

            Returns:
                self: a instância da classe SilverTable.
        """       
        if not self.has_loaded:
            self.load()
            log(__name__).warning(f"The load() method will be called internally because you call create_table first.")        
        self.create_tbl = CreateTableSilver(self.target_df.schema, self.table_name, self.yml).execute()           
        self.create_vw = CreateView(self.layer, self.table_name, self.create_tbl.yml, self.create_tbl.tbl_comment).execute()                
        return self    

    def field(self, source_field_name, field_type=None, target_field_name=None, format_field=True):
        """            
            Método que adiciona de maneira *lazy* as fields que serão consumidas da bronze para a silver.

            Args:
                source_field_name (str): string que representa o nome da field que será consumida da bronze.                
                field_type (str): string que representa o tipo da field que será casteada e formatada para a silver.                
                target_field_name (str): string que representa o nome da field que será adicionada na silver.
                format_field (bool): se verdadeiro formata a field para o tipo especificado.

            Returns:
                self: a instância da classe SilverTable.
        """
        self.dict_of_fields[source_field_name] = (source_field_name, field_type, None, format_field)
        #self.dict_of_fields[source_field_name] = (source_field_name, field_type, target_field_name, format_field)
        return self    




from itertools import combinations
from functools import partial
from dnalib.utils import Utils
from .querytable import QueryTable

###########################################
###### methods by reflection #############
###########################################

class DynamicTable(QueryTable):

    """
        Classe de alto nível para implementar o design pattern reflection baseado nas tabelas do spark. Os métodos são criados dinamicamente a partir das colunas da tabela o que permite uma escrita mais amigável na hora de consultar as tabelas. Basicamente para cada coluna são criados diversos métodos que fazem tarefas simples para filtragem de dados.

        Baseado em: https://stackoverflow.com/questions/29642822/find-name-of-dynamic-method-in-python
        
    """
    def __init__(self, layer, table_name, table_fields=[], select_fields=[], except_fields=[], json_fields=[], query_view_name="view_query_table"):
        """
        """
        super().__init__(query_view_name)
        self.layer = layer
        self.table_name = table_name
        self.source_table = f"{layer}.{table_name}"                
        self.table_fields = table_fields        
        if len(self.table_fields) == 0:
            self.table_fields = Utils.spark_instance().table(self.source_table).columns                
        self.select_fields = select_fields        
        if len(select_fields) == 0:
            self.select_fields = self.table_fields        
        self.except_fields = except_fields        
        self.json_fields = json_fields        
        # campos que serão selecionados
        self.fields = ", ". join([field for field in self.select_fields if field not in self.json_fields])
        # cada combinação vai ser usada para fazer regra "and" e "or" para dois parâmetros
        self.table_fields_combinations = list(combinations(set(self.table_fields)-set(self.json_fields), 2))
        # cria os métodos baseados nas colunas da tabela
        self.reflect()

    def _generate_find_by_first_or_second_parameter(self):
        """
            find_by_{first_field}_or_{second_field}(first_value, second_value)
            find_by_{second_field}_or_{first_field}(second_value, first_value)
        """
        for first_field, second_field in self.table_fields_combinations:                                    
            func = partial(self._find_by_first_or_second_parameter, fields=self.fields, source_table=self.source_table, first_field=first_field, second_field=second_field)
            setattr(self, f"find_by_{first_field.lower()}_or_{second_field.lower()}", func)
            func = partial(self._find_by_first_or_second_parameter, fields=self.fields, source_table=self.source_table, first_field=second_field, second_field=first_field)
            setattr(self, f"find_by_{second_field.lower()}_or_{first_field.lower()}", func)

    def _generate_find_by_first_and_second_parameter(self):
        """
            find_by_{first_field}_and_{second_field}(first_value, second_value)
            find_by_{second_field}_and_{first_field}(second_value, first_value)
        """
        for first_field, second_field in self.table_fields_combinations:                                    
            func = partial(self._find_by_first_and_second_parameter, fields=self.fields, source_table=self.source_table, first_field=first_field, second_field=second_field)
            setattr(self, f"find_by_{first_field.lower()}_and_{second_field.lower()}", func)
            func = partial(self._find_by_first_and_second_parameter, fields=self.fields, source_table=self.source_table, first_field=second_field, second_field=first_field)
            setattr(self, f"find_by_{second_field.lower()}_and_{first_field.lower()}", func)

    def _generate_find_by_single_parameter(self):    
        """
            find_by_{field}(value)
        """
        for field in self.table_fields:     
            if field not in self.json_fields:
                func = partial(self._find_by_single_parameter, fields=self.fields, source_table=self.source_table, field=field)
            else:
                func = partial(self._find_by_json_parameter, fields=self.fields, source_table=self.source_table, json_field=field)
            setattr(self, f"find_by_{field.lower()}", func)

    # outra sintaxe para chamar o método usando apenas um parâmetro
    def find_by_single_parameter(self, **kwargs):
        field_name, field_value = kwargs.popitem()
        return getattr(self, f"find_by_{field_name.lower()}")(field_value)

    def reflect(self):
        # reflect find_all_by        
        self._generate_find_by_single_parameter()
        # reflect find_all_by_and
        self._generate_find_by_first_and_second_parameter()
        # reflect find_all_by_or
        self._generate_find_by_first_or_second_parameter()



from dnalib.log import *
from dnalib.utils import Utils, LandingZoneUtils
from .loader import Loader

class SqlParquetLandingLoader(Loader):
    """ 
        This implementation is based on our adf process:
            table/date/ -> incremental path
            table/date/ -> cdc path
            table/date/ -> full path = if process run in incremental mode, but you specified a full query
            table/ -> full path
    """

    metadata_columns = ["Dt_Inclusao"]

    def __init__(self, source_path, loader_type, table_config):
        super().__init__(source_path, loader_type) 
        self.spark = Utils.spark_instance()  # Obtém a instância do Spark usando Utils.spark_instance()                        
        self.table_config = table_config

    def load_df_parquet(self):
        # le o path mais recente dentro da pasta indicada por source_path
        file_path = LandingZoneUtils.load_last_updated_path(self.source_path)
        # Lê os dados do parquet atual
        try:
            df = self.spark.read.parquet(file_path).drop(*self.metadata_columns) 
        except Exception as e:
            log(__name__).error(f"Error reading parquet file: {e}")          
            raise Exception(f"Error reading parquet file: {e}")
        return df

    def load_incremental(self):       
        return self.load_df_parquet()

    def load_full(self):
        return self.load_df_parquet()
        
class SqlServerParquetLandingLoader(SqlParquetLandingLoader):

    # only applied in cdc purposes
    cdc_metadata_columns = ["LAST_UPDATE", "start_lsn", "end_lsn", "seqval", "operation", "update_mask", "command_id", "data_alteracao"]

    def __init__(self, source_path, loader_type, table_config):
        super().__init__(source_path, loader_type, table_config)

    def load_cdc(self):
        df = self.load_incremental()                

        # query que pega a última alteração de cada registro
        df = (df.withColumnRenamed("__$start_lsn", "start_lsn") 
               .withColumnRenamed("__$end_lsn", "end_lsn") 
               .withColumnRenamed("__$seqval", "seqval") 
               .withColumnRenamed("__$operation", "operation") 
               .withColumnRenamed("__$update_mask", "update_mask") 
               .withColumnRenamed("__$command_id", "command_id"))
        
        # view para filtrar as alterações
        df.createOrReplaceTempView("vw_table_logs")
        self.spark.sql(f"""
            CREATE OR REPLACE TEMP VIEW vw_table_logs_rownumber AS (
                SELECT * FROM (
                    SELECT 
                        *, 
                        ROW_NUMBER() OVER(PARTITION BY {self.table_config['key']} ORDER BY data_alteracao DESC, command_id DESC) AS LAST_UPDATE FROM vw_table_logs
                    WHERE operation <> 3                                                
                )
                WHERE LAST_UPDATE = 1
            );
        """)

        # filtra as operações de insert e update
        df_upinsert = self.spark.sql(""" 
            SELECT * 
            FROM vw_table_logs_rownumber 
            WHERE operation IN (2, 3, 4);
            """).drop(*self.cdc_metadata_columns)                      
        return df_upinsert        

class OracleParquetLandingLoader(SqlParquetLandingLoader):

    def __init__(self, source_path, loader_type, table_config):
        super().__init__(source_path, loader_type, table_config)
        



from dnalib.utils import Utils

def compile_query(query_string, *args, **kwargs):
    """
        Uma função python para compilar as queries usando o método format.
        
        args = [arg_0, arg_1, ..., arg_n]
        kwargs_flatten = [key_0, value_0, key_1, value_1 ..., key_n, value_n]
        kwargs = {key_0:value_0, key_1:value_1 ..., key_n:value_n}
        kwargs_keys = {id_key_0:key_0, id_key_1:key_1 ..., id_key_n:key_n}

        format(arg_0, arg_1, ..., arg_n, key_0, value_0, 
               key_1, value_1 ..., key_n, value_n, 
               {key_0:value_0, key_1:value_1 ..., key_n:value_n,
               id_key_0:key_0, id_key_1:key_1 ..., id_key_n:key_n})

        Usando args:                    
            compile_query("select * from {0}.{1}", "bronze", "pln_ge_pessoa")
                "select * from bronze.pln_ge_pessoa"
        
        Usando kwargs:
            compile_query("select * from {layer}.{table_name}", layer="bronze", table_name="pln_ge_pessoa")
                "select * from bronze.pln_ge_pessoa"

        Usando kwargs sem especificar uma das chaves (usando a posição como em args):
            compile_query("select * from {layer}.{table_name} where {4} = '{5}'", layer="bronze", table_name="pln_ge_pessoa", cpf="12345")
                "select * from bronze.pln_ge_pessoa where cpf = '12345'"

        Usando kwargs acessando a chave e valor dentro da query (a key sempre fica com id_ na frente)
            compile_query("select * from {layer}.{table_name} where {id_cota} = '{cota}'", layer="bronze", table_name="pln_ge_pessoa", cota="134567")
    """              
    # we flat the kwargs to a list with consecutive [key, value, ...]
    kwargs_flatten = [kv_flat for kv in zip(kwargs.keys(), kwargs.values()) for kv_flat in kv]   
    # we create a dictionary with the id_key as key and the key as value
    kwargs_keys = {f"id_{key}" : key for key in kwargs.keys()}
    # merge both dictionaries
    kwargs_full = {**kwargs, **kwargs_keys}
    # finally we compile the query according to a simple logic
    compiled_query = query_string.format(*args, *kwargs_flatten, **kwargs_full)                                               
    return compiled_query

def query(query_string, *args, **kwargs):
    """
        Um decorator para compilar queries em classes ou métodos no formato do spark sql e retornar um dataframe com o resultado.
    """
    def inner(func):          
        def wrapper(*args, **kwargs):                 
            args_idx = 0
            # test if the first parameter is the "self"
            if len(args) > 0 and hasattr(args[0], '__dict__'):                                
                args_idx = 1            
            compiled_query = compile_query(query_string, *args[args_idx:], **kwargs)                                                                
            return func(*args[:args_idx], Utils.spark_instance().sql(compiled_query))
        return wrapper
    return inner

class QueryTable:   

    def __init__(self, query_view_name="view_query_table"):
        self.query_view_name = query_view_name
        self.result_df = None
    
    @query(""" 
        select {fields}
        from {source_table} where {field}::string="{0}" 
    """)
    def _find_by_single_parameter(self, result_df):
        self.result_df = result_df
        self.result_df.createOrReplaceTempView(self.query_view_name)
        return self

    @query(""" 
        select {fields} 
        from {source_table} where {first_field}::string="{0}" and {second_field}::string="{1}" 
    """)    
    def _find_by_first_and_second_parameter(self, result_df):
        self.result_df = result_df
        self.result_df.createOrReplaceTempView(self.query_view_name)
        return self 

    @query(""" 
        select {fields} 
        from {source_table} where {first_field}::string="{0}" or {second_field}::string="{1}" 
    """)
    def _find_by_first_or_second_parameter(self, result_df):
        self.result_df = result_df
        self.result_df.createOrReplaceTempView(self.query_view_name)
        return self 

    @query(""" 
        select {fields}
        from {source_table} where {json_field}:{6}::string="{7}" 
    """)
    def _find_by_json_parameter(self, result_df):
        self.result_df = result_df
        self.result_df.createOrReplaceTempView(self.query_view_name)
        return self

    @query(""" 
        select {fields} from {source_table} 
        where {in_field} in (		    
            select {in_field} from {source_table}	
            where {json_field}:{11}::string="{12}" 
        )
    """)
    def _find_by_json_parameter_in_field(self, result_df):
        self.result_df = result_df
        self.result_df.createOrReplaceTempView(self.query_view_name)
        return self
    


from .loader import SourceType
from .sqlparquetlandingloader import *
from .firebaseloader import *
from dnalib.log import *

class LoaderFactory:
    @staticmethod
    def create_loader(source_type, 
                      source_path, 
                      loader_type, 
                      table_config, 
                      collection_name = None,                      
                      order_by = '__name__', 
                      page_size = 1000, 
                      init_load = None, 
                      end_load = None,
                      flat_json = True,
                      db_name = None,
                      spark_num_partitions = None):
        if source_type.value == SourceType.SQLSERVER.value:
            return SqlServerParquetLandingLoader(source_path, loader_type, table_config)
        elif source_type.value == SourceType.ORACLE.value:
            return OracleParquetLandingLoader(source_path, loader_type, table_config)
        elif source_type.value == SourceType.FIREBASE.value:
            return FireBaseLoader(source_path, loader_type, collection_name, order_by, page_size, init_load, end_load, flat_json, db_name, spark_num_partitions)
        else:
            log(__name__).error(f"file_loader could not be created for source_type {source_type}")
            raise ValueError(f"file_loader could not be created for source_type {source_type}")


class Field:

    def __init__(self, 
                 field_name,                  
                 field_comment=None,                                  
                 field_type=None,
                 field_is_nullable=None, 
                 field_format=None):   

        self.field_name = field_name    
        # dataCarga is not required in yml, cause it is inserted by code            
        if self.field_name == "dataCarga":
            self.field_comment = "Data de carga dos dados"
        # same as dataCarga            
        elif self.field_name == "checksum":
            self.field_comment = "CheckSum/Hash_sql das informações do registro (linha)"
        elif field_comment:
            self.field_comment = field_comment
        else:
            self.field_comment = ""
        if field_type:
            self.field_type = field_type
        else:
            self.field_type = 'string'
        # field is only declared as not nulabble if it is false
        if field_is_nullable != False:
            self.field_is_nullable = True 
        else:
            self.field_is_nullable = False
        # formatation is not aplied only if it is false
        if field_format != False:
            self.field_format = True 
        else:
            self.field_format = False

    def is_valid_comment(self):
        return self.comment != ""

    def hash_sql(self):
        return f"sha2(cast({self.field_name} as string), 256) AS {self.field_name}"

    def to_sql(self):
        sql_representation = "{} {} COMMENT '{}'{}".format(self.field_name, self.field_type, self.field_comment, " NOT NULL" if not self.field_is_nullable else "")  
        return sql_representation
    
    def __repr__(self):
        return self.to_sql()
    
    def __str__(self):
        return f"""Field(field_name={self.field_name}, field_type={self.field_type}, field_comment={self.field_comment}, field_is_nullable={self.field_is_nullable}, field_format={self.field_format})"""
    


from pyspark.sql.functions import lit
from pyspark.sql.types import StructType, StringType, IntegerType, DateType, StructField, TimestampType
from dnalib.writer import BatchTableWriter
from dnalib.core import Table
from dnalib.tables import DynamicTable
from dnalib.log import *
from .block import Block
from .key import Key

class BlockBucket(Table,
                  BatchTableWriter,
                  DynamicTable):
    """
        Classe que controla os "buckets", ou seja os baldes de blocos.
    """
    
    fields_comment = {
        "is_block_processed": {'comment': "Flag que indica se o bloco já foi processado, ou seja carregado em alguma blockchain."}
    }

    schema = (StructType([field for field in Block.schema])
                    .add(StructField("is_block_processed", IntegerType(), True, fields_comment["is_block_processed"])))
    
    layer = "blockchain"

    def __init__(self, table_name):    
        # chamando o super para criar a tabela
        Table.__init__(self, self.layer, table_name, self.schema)          
        BatchTableWriter.__init__(self, self.layer, table_name)        
        self.create()
        # chamando o super para a DynamicTable
        DynamicTable.__init__(self, self.layer, table_name, json_fields=["payload"])      

    def add_block_metadata(self, block, prev_block_name=None, prev_block_join_rules=None):
        # the metadata is necessary to translate rules to blockchain
        tbl_properties = {f"blockchain.{block.block_name}.prev_block_name": prev_block_name, 
                          f"blockchain.{block.block_name}.prev_block_join_rules": prev_block_join_rules.serialize() if prev_block_join_rules is not None else None}
        # we store metadata as tbl_properties
        self.add_tbl_properties(tbl_properties)

    def add_block(self, block, prev_block_name=None, prev_block_join_rules=None):        
        # add block metadata
        self.add_block_metadata(block.load(), prev_block_name, prev_block_join_rules)
        # adding field to control if the block is already in the blockchain
        df_block = block.df_block.withColumn("is_block_processed", lit(0))
        # inserting block for only new data (ignoring updates)
        log(__name__).info(f"Appeding new blocks to bucket table {self.layer}.{self.table_name}.")
        self.upsert(df_block, "source.id_block == update.id_block", ignore_updates=True)
        # get operation metrics to see new blocks
        total_new_blocks = self.parse_last_operation_metrics().get("numTargetRowsInserted", 0)
        log(__name__).info(f"Total of {total_new_blocks} blocks of type {block.block_name} were added to bucket.")

    def load(self, block_name):    
        # load block
        df_block = self.spark.table(self.table_catalog_path).filter(f"block_name == '{block_name}' and is_block_processed = 0").drop("is_block_processed")
        # get metadata
        tbl_properties = self.parse_tbl_properties()
        # load block metadata
        prev_block_name = tbl_properties.get(f"blockchain.{block_name}.prev_block_name", None)
        prev_block_join_rules = tbl_properties.get(f"blockchain.{block_name}.prev_block_join_rules", None)
        # loading join rules
        if prev_block_join_rules is not None:
            prev_block_join_rules = Key.from_serialized(prev_block_join_rules)
        # retorna o bloco carregado
        return df_block, prev_block_name, prev_block_join_rules
            


from pyspark.sql.functions import xxhash64, col, lit, concat, from_utc_timestamp, current_timestamp, to_json, struct
from pyspark.sql.types import StructType, StringType, IntegerType, DateType, StructField, TimestampType

class Block:    
    """ 
        A classe Block representa um bloco na cadeia do Blockchain.

        Args:
            df (DataFrame): dataframe que representa os dados que serão adicionados a blockchain.
            block_name (str): nome do bloco.
            block_business_keys (list): chaves de negócio que são usadas pra conectar os blocos.      
            fields_to_hash (list, default=[]): lista de campos que serão usados para gerar o hash identificador do bloco. se vazia, todos os campos são usados na geração do hash.
            fields_to_non_hash (list, default=[]): lista de campos que serão ignorados para gerar o hash identificador do bloco.

        Attributes:
            fields_ignored_to_hash (list): lista que contém os campos que são ignorados por padrão para gerar o hash identificador do bloco.
            df (DataFrame): dataframe que representa os dados que serão adicionados a blockchain.
            block_name (str): nome do bloco.
            block_business_keys (list): chaves de negócio que são usadas pra conectar os blocos.      
            fields_to_hash (list, default=[]): lista de campos que serão usados para gerar o hash identificador do bloco. se vazia, todos os campos são usados na geração do hash.
            fields_to_non_hash (list, default=[]): lista de campos que serão ignorados para gerar o hash identificador do bloco.
            df_block (DataFrame): dataframe que contém a estrutura inicial do bloco.
    """

    fields_comment = {
        "block_name":  {'comment':"Nome do bloco, utilizado para otimizar as conexões do blockchain."},
        "block_level":  {'comment':"Nível do bloco, usado para identificar a sua profundidade na blockchain."},
        "block_business_keys":  {'comment':"Campo com as chaves de negócio concatenadas."},
        "id_block_gen":  {'comment':"Identificador gerado no primeiro bloco e propagado por toda a cadeia."},
        "id_block":  {'comment':"Identificador único do bloco, gerado a partir de um hash."},
        "id_block_prev":  {'comment':"Identificador id_block do bloco anterior."},
        "is_gen":  {'comment':"Flag que indica se o bloco é gerador (1 se sim e 0 se não)."},
        "block_timestamp":  {'comment':"Timestamp da inserção do bloco na blockchain."},
        "payload":  {'comment':"Metadados em formato JSON, são as colunas do dataframe que representa o bloco."},
        "last_block":  {'comment':"Flag que indica que o bloco é o último na cadeia."},
    }

    schema = StructType([
        StructField("block_name", StringType(), True, fields_comment["block_name"]),
        StructField("block_level", IntegerType(), True, fields_comment["block_level"]),
        StructField("block_business_keys", StringType(), True, fields_comment["block_business_keys"]),
        StructField("id_block_gen", StringType(), True, fields_comment["id_block_gen"]),
        StructField("id_block", StringType(), True, fields_comment["id_block"]),
        StructField("id_block_prev", StringType(), True, fields_comment["id_block_prev"]),
        StructField("is_gen", IntegerType(), True, fields_comment["is_gen"]),
        StructField("block_timestamp", TimestampType(), True, fields_comment["block_timestamp"]),
        StructField("payload", StringType(), True, fields_comment["payload"]),
        StructField("last_block", IntegerType(), True, fields_comment["last_block"]),  
    ])

    fields_ignored_to_hash = ["dataCarga", "checksum", "data_carga"]

    def __init__(self, 
                 df, 
                 block_name,    
                 block_business_keys,             
                 fields_to_hash=[], 
                 fields_to_non_hash=[]):
        self.df = df
        self.block_name = block_name       
        self.block_business_keys = block_business_keys 
        self.fields_to_hash = fields_to_hash
        if len(self.fields_to_hash) == 0:            
            self.fields_to_hash = self.df.columns
        self.fields_to_non_hash = fields_to_non_hash + self.fields_ignored_to_hash
        self.df_block = None
    
    def hash(self):
        hash_fields = sorted(set(self.fields_to_hash) - set(self.fields_to_non_hash))
        return xxhash64(*hash_fields)
    
    def concat_business_keys(self):
        list_of_end_keys = []
        for i, key in enumerate(self.block_business_keys):        
            separator = [lit("|")] if i < len(self.block_business_keys)-1 else []                
            list_of_end_keys = list_of_end_keys + [col(key)] + separator
        return concat(*list_of_end_keys)

    def load(self):
        # load data to dataframe
        self.df_block = (self.df.select(
            lit(self.block_name).alias("block_name"),
            lit(0).alias("block_level"),
            self.concat_business_keys().alias("block_business_keys"),
            self.hash().alias("id_block_gen"),
            self.hash().alias("id_block"), 
            lit(None).alias("id_block_prev"), 
            lit(1).alias("is_gen"),             
            from_utc_timestamp(current_timestamp(), "America/Sao_Paulo").alias("block_timestamp"),     
            to_json(struct("*")).alias("payload"),
            lit(1).alias("last_block")
        ))
        return self
        



from dnalib.utils import TableUtils, Utils
from dnalib.log import log

class TableComment:             
    """
        Classe construída para compilar o template do comentário da tabela, seguindo os critérios do catalogo do Atlan.

        Args:
            layer (str): camada da tabela no lake.   
            table_name (str): string que representa o nome da tabela.
            yml (yml object or str, optional): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é None.
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.

        Attributes:
            table_comment_atlan_pattern (dict): dicionário com as chaves para o comentário da tabela, seguindo os critérios do catalogo do Atlan.
            yml (yml object ou None): instância do objeto yml, é None caso nenhum arquivo seja informado.
            layer (str): camada da tabela no lake.   
            table_name (str): string que representa o nome da tabela.
            anonimized_fields (list): lista de campos para serem anonimizados.
            comment (dict): dicionário com o padrão table_comment_atlan_pattern.key.
            comment_squad (dict): dicionário com o padrão table_comment_atlan_pattern.key.
            parsed_comment (str): string com o template do comentário da tabela compilado, preenchido após a chamada do método parse.
            comment_complete (dict): merge entre os dicionários comment e comment_squad, preenchido após a chamada do método parse.
            parsed_template (str): template do comentário da tabela compilado, preenchido após a chamada do método parse.
    """ 

    table_comment_keywords = {
        'descricao': '- **Descrição**:',
        'sistema_origem': '- **Sistema Origem**:',
        'calendario': '- **Calendário de Disponibilização**:',
        'tecnologia': '- **Tecnologia Sistema Origem**:',
        'camada': '- **Camada de Dados**:',
        'peridiocidade': '- **Periodicidade de Atualização**:',
        'retencao': '- **Período de Retenção**:',
        'vertical': '- **Vertical**:',
        'dominio': '- **Domínio**:',
        'squad': '- **Squad Responsável**:',
        'categoria_portosdm': '- **Categoria e Grupo PortoSDM**:',
        'confidencialidade': '- **Classificação da Confidencialidade**:',
        'classificacao': '- **Classificação**:',
        'campos_anonimizados': '- **Campos Anonimizados**:',
        'deprecated': '- **Deprecated**:'
    }

    table_squad_keywords = {
        'gi': '- **GI - Gestor da Informação**:',
        'gdn': '- **GDN - Gestor de Negócio**:',
        'curador': '- **Curador**:',
        'custodiante': '- **Custodiante**:',
    }

    table_comment_atlan_pattern =  {
        'descricao': '- **Descrição**:',
        'sistema_origem': '- **Sistema Origem**:',
        'calendario': '- **Calendário de Disponibilização**:',
        'tecnologia': '- **Tecnologia Sistema Origem**:',
        'camada': '- **Camada de Dados**:',
        'peridiocidade': '- **Periodicidade de Atualização**:',
        'retencao': '- **Período de Retenção**:',
        'vertical': '- **Vertical**:',
        'dominio': '- **Domínio**:',
        'gi': '- **GI - Gestor da Informação**:',
        'gdn': '- **GDN - Gestor de Negócio**:',
        'curador': '- **Curador**:',
        'custodiante': '- **Custodiante**:',
        'squad': '- **Squad Responsável**:',
        'categoria_portosdm': '- **Categoria e Grupo PortoSDM**:',
        'confidencialidade': '- **Classificação da Confidencialidade**:',
        'classificacao': '- **Classificação**:',
        'campos_anonimizados': '- **Campos Anonimizados**:',
        'deprecated': '- **Deprecated**:'
    }

    def __init__(self, 
                 layer, 
                 table_name, 
                 yml={}, 
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}):
        self.yml = yml
        # yml must be either a string or a dict from the yml file
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.table_name = table_name.strip().lower()
        self.layer = layer.strip().lower()   
        self.anonimized_fields = self.yml.get("anonimized_fields", anonimized_fields)
        self.comment = self.yml.get("comment", comment)
        self.comment_squad = self.yml.get("comment_squad", comment_squad)
        self.parsed_comment = ""
        self.comment_complete = {}
        self.parsed_template = ""

    def parse_parameters_from_table_comment(self):
        """ 
            Método interno para tentar parsear a string do comentário da tabela.
        """
        parsed_str_comment = TableUtils.table_comment(self.layer, self.table_name)
        list_comment_pattern_clean = []

        # in this solution we consider that each keyword is separated by '\n'
        for table_content in parsed_str_comment.split('\n'):
            # remove empty string
            comment_patern_clean = table_content.strip()
            if comment_patern_clean != '':
                list_comment_pattern_clean.append(comment_patern_clean)

        comment = dict.fromkeys(self.table_comment_keywords.keys(), "")
        comment_squad = dict.fromkeys(self.table_squad_keywords.keys(), "")

        # for each keyword we try to find its definition in the list of the table comment
        for table_comment_key in self.table_comment_atlan_pattern:
            table_keyword = self.table_comment_atlan_pattern[table_comment_key]    
            for comment_patern_clean in list_comment_pattern_clean:                
                comment_idx = comment_patern_clean.find(":")+1        
                if table_keyword.upper() in comment_patern_clean.upper() and comment_idx < len(comment_patern_clean):    
                    if table_comment_key in comment:        
                        comment[table_comment_key] = comment_patern_clean[comment_idx:].strip()
                    else:
                        comment_squad[table_comment_key] = comment_patern_clean[comment_idx:].strip()  

        # try to load anonimized fields from the table comment
        str_anonimized_fields = comment["campos_anonimizados"]       
        list_anonimized_fields = str_anonimized_fields.replace(".", "").replace(" e ", ",").split(",")  
        anonimized_fields = [] 
        for anonimized_field in list_anonimized_fields:
            anonimized_field_clean = anonimized_field.strip()
            if anonimized_field_clean != '':
                anonimized_fields.append(anonimized_field_clean.lower())

        return comment, comment_squad, anonimized_fields

    def __load_and_verify_params(self):
        """
            Método interno que carrega as estruturas principais para popular os comentários da tabela.

            Raises:
                ValueError: Ocorre se alguma das estruturas comment ou comment_squad não forem informadas. Ou devem ser passadas por parâmetros, ou devem estar dentro do yml.
        """                
        if len(self.yml) == 0:
            (comment, comment_squad, anonimized_fields) = self.parse_parameters_from_table_comment()
            if len(self.comment) == 0:  
                self.comment = comment              
            if len(self.comment_squad) == 0:
                self.comment_squad = comment_squad
            if len(self.anonimized_fields) == 0:
                self.anonimized_fields = anonimized_fields
    
        # merge both dicts
        self.comment_complete = {**self.comment, **self.comment_squad}
        self.comment_complete["campos_anonimizados"] = ", ".join(self.anonimized_fields)
        # deprecated can be or not informed
        self.comment_complete["deprecated"] = self.comment_complete.get("deprecated", "")
    
    def parse(self): 
        """
            Método que faz o parse dos comentários baseados na estrutura do Atlan. Os valores estão no dicionário TableComment.table_comment_atlan_pattern.

            Returns:
                self: a instância da classe TableComment.

            Raises:
                ValueError: Se a estrutura do comentário não for informada corretamente (ou seja, se faltar alguma chave).
        """        
        self.__load_and_verify_params()         
        for key in self.table_comment_atlan_pattern:
            if not key in self.comment_complete:
                log(__name__).error(f"You must inform the parameter {key} in table comment.")
                raise ValueError(f"You must inform the parameter {key} in table comment.")
            else:                
                self.parsed_comment += f"{self.table_comment_atlan_pattern[key]} {TableUtils.format_comment_content(self.comment_complete[key])}\n"
        return self
    
    def template(self):             
        """
            Método que constroí o template da estrutura do comentário da tabela baseado nos parâmetros informados.

            .. code-block:: SQL

                COMMENT ON TABLE layer.table_name IS "comentário compilado"
            
            Returns:
                self: a instância da classe TableComment.
        """   

        self.parse()    
        self.parsed_template = """
            COMMENT ON TABLE {}.{} IS "{}"
        """.format(self.layer, self.table_name, self.parsed_comment)           

        return self             

    def execute(self):        
        """
            Método que executa todo processo para gerar e escrever o comentário da tabela.              
            
            Returns:                
                self: a instância da classe TableComment.

            Raises:
                Exception: Se a tabela informada não existir.
        """

        # verify if table exists
        if not TableUtils.table_exists(self.layer, self.table_name):
            log(__name__).error(f"Table {self.layer}.{self.table_name} does not exist")
            raise Exception(f"Table {self.layer}.{self.table_name} does not exist")

        # generate final template
        self.template()

        # run comment on table
        Utils.spark_instance().sql(self.parsed_template)        
        return self



from dnalib.utils import TableUtils, Utils
from dnalib.log import *

class DDL:    
    """ """    
    def __init__(self, layer):
        self.layer = layer.strip().lower()   
        self.spark = Utils.spark_instance() 
        self.dbutils = Utils.dbutils()

class TableDDL(DDL):
    """ """
    def __init__(self, layer, table_name):
        DDL.__init__(self, layer)          
        self.table_name = table_name.strip().lower()
        self.table_path = Utils.lakehouse_path(self.layer, self.table_name)
        self.table_catalog_path = f"{self.layer}.{self.table_name}"

    def describe(self):
        raise NotImplementedError("Method describe() must be implemented.")    

    def create(self):
        raise NotImplementedError("Method create() must be implemented.")    

    def create_view(self):
        raise NotImplementedError("Method create_view() must be implemented.")    
        
    def drop(self):
        raise NotImplementedError("Method drop() must be implemented.")        

    def drop_view(self):
        raise NotImplementedError("Method drop_view() must be implemented.")      

    def exists(self):
        raise NotImplementedError("Method exists() must be implemented.")        

    def view_exists(self):
        raise NotImplementedError("Method view_exists() must be implemented.")        



from dnalib.log import *
from enum import Enum

class SourceType(Enum):
    """ 
        Enum que define os métodos validos para origem de dados: TXT, CSV, PARQUET, KAFKA e EVENTHUB (WIP); SQLSERVER, ORACLE (implementados).
    """
    TXT = "txt"
    CSV = "csv"
    PARQUET = "parquet"    
    KAFKA = "kafka"
    EVENTHUB = "eventhub"
    SQLSERVER = "sqlserver"
    ORACLE = "oracle"
    FIREBASE = "firebase"

class LandingLoader(Enum):
    """ 
        Enum que define a forma como os dados serão carregados a partir da origem: 
        INCREMENTAL: considera que a origem é um arquivo incremental e define automaticamente que será feito um upsert.
        FULL: considera que a origem é um arquivo completo, e define automaticamente que será feito um overwrite.
        CDC: considera que a origem é um arquivo CDC (funciona somente para SourceType.SQLSERVER) e define automaticamente que será feito um upsert.
        STREAMING: WIP.
    """
    INCREMENTAL = "incremental" # read a single part, and merge data
    FULL = "full" # read all, then overwrite    
    CDC = "cdc" # only for cdc loader    
    STREAMING = "streaming" # for streaming purpose

class Loader:
    """
        A classe Loader é uma implementação de alto nível que define a carga de um arquivo da landing zone (todas ingestões batch são baseadas em arquivos
        seja csv, txt ou banco, que nesse último caso usa parquet).

        Args:
            source_path (str): caminho na landingzone de onde o arquivo será lido.
            table_name (str): string que representa o nome da tabela.          
            loader_type (LandingLoader): algum dos tipos válidos da LandingLoader.            
            name_filter (str, optional): utilizado caso seja necessário filtrar arquivos dentro de um mesmo diretório da landing.

        Attributes:
            source_path (str): caminho na landingzone de onde o arquivo será lido.
            table_name (str): string que representa o nome da tabela.    
            loader_type (LandingLoader): algum dos tipos válidos da LandingLoader.            
            name_filter (str, optional): utilizado caso seja necessário filtrar arquivos dentro de um mesmo diretório da landing.

    """
    def __init__(self, source_path, loader_type):
        self.source_path = source_path        
        self.loader_type = loader_type                
        self.df_loader = None

    def load_incremental(self):
        # TODO: implementar o load do tipo incremental. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_incremental() must be implemented.")

    def load_full(self):
        # TODO: implementar o load do tipo full. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_full() must be implemented.")

    def load_cdc(self):
        # TODO: implementar o load do tipo cdc. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_cdc() must be implemented.")    

    def load(self):
        if self.loader_type.value == LandingLoader.INCREMENTAL.value:
            self.df_loader = self.load_incremental()
        elif self.loader_type.value == LandingLoader.FULL.value:
            self.df_loader = self.load_full()
        elif self.loader_type.value == LandingLoader.CDC.value:
            self.df_loader = self.load_cdc()
        else:
            log(__name__).error(f"Invalid load type {self.loader_type}")
            raise ValueError(f"Invalid load type {self.loader_type}")
        log(__name__).info(f"Load of {self.loader_type.value} type completed successfully.")
        return self



from dnalib.utils import Utils
from dnalib.log import *
from dnalib.loaders import LoaderFactory, SourceType, LandingLoader
from dnalib.core import CreateTableBronze
from dnalib.writer import WriteModes
from .tables import LayerTable
        
class BronzeTable(LayerTable):

    layer = "bronze"

    def __init__(self,                  
                 table_name,                 
                 source_path,
                 yaml_config_path,
                 table_sufix = "",
                 source_type=SourceType.SQLSERVER, 
                 loader_type=LandingLoader.FULL,
                 source_df=None, 
                 name_filter="", 
                 unzip=True, 
                 include_checksum=False,  
                 collection_name=None,               
                 order_by = '__name__',
                 page_size = 1000,
                 init_load = None, 
                 end_load = None,
                 flat_json = True,
                 db_name = None,
                 spark_num_partitions = None):
        super().__init__(self.layer, f"{table_sufix}_{table_name}", source_df, include_checksum)        
        self.source_path = source_path
        self.name_filter = name_filter
        self.table_sufix = table_sufix
        self.source_type = source_type
        self.yaml_config_path = yaml_config_path
        self.table_config = Utils.yaml_table_parameters(table_name=f"{table_sufix}_{table_name}", yml_file_path=yaml_config_path)
        self.loader_type = loader_type
        self.unzip = unzip      
        self.collection_name = collection_name     
        self.order_by = order_by
        self.page_size = page_size
        self.init_load = init_load
        self.end_load = end_load  
        self.flat_json = flat_json
        self.db_name = db_name
        self.spark_num_partitions = spark_num_partitions
        self.except_fields = {}

        # Usa a fábrica para criar o file_loader correto
        self.file_loader = LoaderFactory.create_loader(            
            self.source_type,
            self.source_path,
            self.loader_type,            
            self.table_config,     
            self.collection_name,       
            self.order_by,
            self.page_size,
            self.init_load,
            self.end_load, 
            self.flat_json,
            self.db_name,
            self.spark_num_partitions
        )        

    def parse_df_source(self):
        """
            Método que executa o carregamento de dados da camada landing para a bronze

            Returns:
                source_df (spark DataFrame): dataframe carregado a partir da camada source, caso source_df seja None.

        """         
        # load strategy in this case is to check if the table exists        
        return self.file_loader.load().df_loader

    def load_fields_from_source(self):                
        return "*"
    
    def create_table(self, yml={}, partition_fields=[], tbl_properties={}, replace=False):
        if not self.has_loaded:
            self.load()
            log(__name__).warning(f"The load() method will be called internally because you call create_table first.") 
        self.creat_tbl = CreateTableBronze(self.target_df.schema, self.table_name, yml, partition_fields, tbl_properties, replace).execute()
        return self
    
    # implementar lógica usando mode = None (nesse caso a decisão usa o LandingLoader como parâmetro caso contrário fica a critério de quem escolheu)
    def persist(self, mode=None, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None):        
        has_delete = False                
        # when you have cdc ingestion, you may need to upsert
        if mode is None:
            if self.loader_type.value == LandingLoader.CDC.value:
                mode = WriteModes.UPSERT
                merge_condition = self.table_config["merge"]
                has_delete = True        
            elif self.loader_type.value == LandingLoader.INCREMENTAL.value:
                mode = WriteModes.UPSERT
                merge_condition = self.table_config["merge"]  
            elif self.loader_type.value == LandingLoader.FULL.value:
                mode = WriteModes.OVERWRITE      

        # calling super persist method
        super().persist(mode=mode, partition_fields=partition_fields, optimize=optimize, source_df_name=source_df_name, update_df_name=update_df_name, merge_condition=merge_condition)

        # only for cdc purposes
        if has_delete:            
            key_columns = ",".join(self.table_config["key"]) if isinstance(self.table_config["key"], list) else self.table_config["key"]
            delete_condition = f"SELECT concat({key_columns}) FROM vw_table_logs_rownumber WHERE operation = 1"
            self.writer.delete(key_columns, delete_condition)       
        return self     




from dnalib.writer import BatchTableWriter
from dnalib.utils import TableUtils, Utils
from dnalib.log import *
from .ddl import TableDDL
from .field import Field

class Table(TableDDL):

    keywords = {
        'descricao': '- **Descrição**:',
        'sistema_origem': '- **Sistema Origem**:',
        'calendario': '- **Calendário de Disponibilização**:',
        'tecnologia': '- **Tecnologia Sistema Origem**:',
        'camada': '- **Camada de Dados**:',
        'peridiocidade': '- **Periodicidade de Atualização**:',
        'retencao': '- **Período de Retenção**:',
        'vertical': '- **Vertical**:',
        'dominio': '- **Domínio**:',
        'squad': '- **Squad Responsável**:',
        'categoria_portosdm': '- **Categoria e Grupo PortoSDM**:',
        'confidencialidade': '- **Classificação da Confidencialidade**:',
        'classificacao': '- **Classificação**:',
        'campos_anonimizados': '- **Campos Anonimizados**:',
        'deprecated': '- **Deprecated**:'
    }

    squad_keywords = {
        'gi': '- **GI - Gestor da Informação**:',
        'gdn': '- **GDN - Gestor de Negócio**:',
        'curador': '- **Curador**:',
        'custodiante': '- **Custodiante**:'
    }

    atlan_keywords =  {
        'descricao': '- **Descrição**:',
        'sistema_origem': '- **Sistema Origem**:',
        'calendario': '- **Calendário de Disponibilização**:',
        'tecnologia': '- **Tecnologia Sistema Origem**:',
        'camada': '- **Camada de Dados**:',
        'peridiocidade': '- **Periodicidade de Atualização**:',
        'retencao': '- **Período de Retenção**:',
        'vertical': '- **Vertical**:',
        'dominio': '- **Domínio**:',
        'gi': '- **GI - Gestor da Informação**:',
        'gdn': '- **GDN - Gestor de Negócio**:',
        'curador': '- **Curador**:',
        'custodiante': '- **Custodiante**:',
        'squad': '- **Squad Responsável**:',
        'categoria_portosdm': '- **Categoria e Grupo PortoSDM**:',
        'confidencialidade': '- **Classificação da Confidencialidade**:',
        'classificacao': '- **Classificação**:',
        'campos_anonimizados': '- **Campos Anonimizados**:',
        'deprecated': '- **Deprecated**:'
    }

    atlan_non_required_keywords = ['campos_anonimizados', 'deprecated']

    def __init__(self, 
                 layer, 
                 table_name, 
                 schema=None,
                 yml={}, 
                 partition_fields=[],    
                 anonimized_fields=[],                 
                 tbl_properties={},
                 comment={}, 
                 comment_squad={},          
                 fields={},                 
                 replace=False):
        TableDDL.__init__(self, layer, table_name)                            
        self.schema = schema
        self.yml = yml
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.partition_fields = self.yml.get("partition_fields", partition_fields)  
        self.anonimized_fields = self.yml.get("anonimized_fields", anonimized_fields)    
        self.tbl_properties = self.yml.get("tbl_properties", tbl_properties)  
        self.comment = self.yml.get("comment", comment)    
        self.comment_squad = self.yml.get("comment_squad", comment_squad)    
        self.fields = self.yml.get("fields", fields)                  
        self.fields_metadata = []
        self.fields_catalog_information = None    
        self.replace = replace     
        self.comment_complete = {}
        self.parsed_table = ""
        self.parsed_view = ""
        self.sql_parsed_comment = ""
        self.sql_fields_metadata = "" 
        self.sql_tbl_properties = "" 
        self.sql_partition_fields = ""    
        self.sql_parsed_comment = "" 
        self.last_operation_metrics = {}

    # validation methods (for advanced tables purpores)
    def validate_comment_complete(self):
        """ 
            Método que valida se os comentários estão seguindo o padrão exigido pelo Atlan, o conteúdo não é validado, apenas a estrutura.
        """
        # parse comment complete structure
        self.parse_comment_complete()
        # validate if it has all atlan keywords
        diff_atlan_keywords = set(self.atlan_keywords.keys()) - set(self.comment_complete.keys())
        # if there is some difference, then some keywords were not informed
        if len(diff_atlan_keywords) > 0:
            log(__name__).error(f"You must inform all comment keywords: {diff_atlan_keywords}")
            raise Exception(f"You must inform all comment keywords: {diff_atlan_keywords}")
        # now we validate all values except anonimized fields and deprecated (they are not required)        
        required_comments = {key:value for key, value in self.comment_complete.items() if key not in Table.atlan_non_required_keywords}        
        # if some comment is empty we do not accept it
        if not all(required_comments.values()):
            empty_keys = [key for key, value in required_comments.items() if len(value) == 0]
            log(__name__).error(f"Those keys are empty and must be informed: {empty_keys}")
            raise Exception(f"Those keys are empty and must be informed: {empty_keys}")

    # parse methods
    def parse_last_operation_metrics(self):
        df_desc_history = TableUtils.describe_history(self.layer, self.table_name, limit=1)
        last_operation_metrics = df_desc_history.select("operationMetrics").collect()[0].asDict()
        self.last_operation_metrics = last_operation_metrics.get("operationMetrics")
        return self.last_operation_metrics

    def parse_parameters_from_table_comment(self):
        parsed_str_comment = TableUtils.table_comment(self.layer, self.table_name)
        list_comment_pattern_clean = []

        # in this solution we consider that each keyword is separated by '\n'
        for table_content in parsed_str_comment.split("\n"):
            # remove empty string
            comment_pattern_clean = table_content.strip()
            if comment_pattern_clean != '':
                list_comment_pattern_clean.append(comment_pattern_clean)

        comment = dict.fromkeys(self.keywords.keys(), "")
        comment_squad = dict.fromkeys(self.squad_keywords.keys(), "")

        # for each keyword we try to find its definition in the list of the table comment
        for table_comment_key in self.atlan_keywords:
            table_keyword = self.atlan_keywords[table_comment_key]    
            for comment_pattern_clean in list_comment_pattern_clean:                
                comment_idx = comment_pattern_clean.find(":")+1        
                if table_keyword.upper() in comment_pattern_clean.upper() and comment_idx < len(comment_pattern_clean):    
                    if table_comment_key in comment:        
                        comment[table_comment_key] = comment_pattern_clean[comment_idx:].strip()
                    else:
                        comment_squad[table_comment_key] = comment_pattern_clean[comment_idx:].strip()  

        # try to load anonimized fields from the table comment
        str_anonimized_fields = comment["campos_anonimizados"]       
        list_anonimized_fields = str_anonimized_fields.replace(".", "").replace(" e ", ",").split(",")  
        anonimized_fields = [] 
        for anonimized_field in list_anonimized_fields:
            anonimized_field_clean = anonimized_field.strip()
            if anonimized_field_clean != '':
                anonimized_fields.append(anonimized_field_clean.lower())

        return comment, comment_squad, anonimized_fields

    def parse_comment_complete(self):
        """ """
        if len(self.comment_complete) == 0:
            if len(self.yml) == 0 and self.exists():
                (comment, comment_squad, anonimized_fields) = self.parse_parameters_from_table_comment()
                if len(self.comment) == 0:  
                    self.comment = comment              
                if len(self.comment_squad) == 0:
                    self.comment_squad = comment_squad
                if len(self.anonimized_fields) == 0:
                    self.anonimized_fields = anonimized_fields
    
            # merge both dicts
            self.comment_complete = {**self.comment, **self.comment_squad}       

            # adjust non required parameters     
            self.comment_complete["campos_anonimizados"] = ", ".join(self.anonimized_fields)
            self.comment_complete.setdefault("deprecated", "")            
        return self.comment_complete

    def parse_fields_catalog_information(self):
        """ """
        if self.fields_catalog_information is None and self.exists():
            self.fields_catalog_information = self.spark.catalog.listColumns(f"{self.layer}.{self.table_name}")
        return self.fields_catalog_information    

    def parse_tbl_properties(self): 
        """ """                 
        if len(self.tbl_properties) == 0 and self.exists():
            df_tbl_properties = self.show_tbl_properties()
            self.tbl_properties = {tbl_property[0]:tbl_property[1] for tbl_property in df_tbl_properties.collect()}            
        return self.tbl_properties        

    def parse_partition_fields(self):
        """ """
        if len(self.partition_fields) == 0 and self.exists():            
            self.parse_fields_catalog_information()
            for field in self.fields_catalog_information:
                if field.isPartition:
                    self.partition_fields.append(field.name)
        return self.partition_fields

    def parse_fields_metadata(self):
        """ """                
        if len(self.fields_metadata) == 0:
            if len(self.fields) == 0 and self.exists():                                                    
                self.parse_fields_catalog_information()
                for field in self.fields_catalog_information:
                    self.fields_metadata.append(Field(field.name, field.description, field.dataType, field.nullable))
            else:                
                if self.schema != None:                                    
                    for field in self.schema:     
                        field_params = self.fields.get(field.name, [])
                        # comment can only comes for yml (currently)
                        field_comment = (field_params[0:1] or [field.metadata.get("comment", None)])[0]                                                  
                        # if field comes from yml, we ignore datatype of schema, because it can be a parameter to cast data
                        field_type = (field_params[1:2] or [None])[0] or field.dataType.simpleString()
                        # we only consider nullable from yml if it is True, otherwise the value is keep from schema
                        field_is_nullable = (field_params[2:3] or [None])[0] or field.nullable
                        # format can only comes from yml
                        field_format = (field_params[3:4] or [None])[0]
                        self.fields_metadata.append(Field(field.name, field_comment, field_type, field_is_nullable, field_format))        
                else:
                    for field_name, field_params in self.fields.items():                             
                        self.fields_metadata.append(Field(field_name, *field_params))        
        return self.fields_metadata
    
    def table_comment_to_sql(self):
        """ """
        self.parse_comment_complete()
        # parameters "campos_anonimizados" and "deprecated" are not required
        if len(self.comment_complete) > 2:
            list_of_comments = [f"{self.atlan_keywords[key]} {TableUtils.format_comment_content(self.comment_complete[key])}" for key in self.comment_complete]
            self.sql_parsed_comment = "\n".join(list_of_comments)                        
        return self.sql_parsed_comment

    def fields_metadata_to_sql(self):
        """ """
        self.parse_fields_metadata()
        if len(self.fields_metadata) > 0:            
            self.sql_fields_metadata = str(self.fields_metadata)[1:-1]
        return self.sql_fields_metadata

    def tbl_properties_to_sql(self):
        """ """
        self.parse_tbl_properties()
        if len(self.tbl_properties) > 0:
            tbl_properties_content = ", ".join([f"{key}='{value}'" for key, value in self.tbl_properties.items()])            
            self.sql_tbl_properties = f"TBLPROPERTIES ({tbl_properties_content})"
        return self.sql_tbl_properties
    
    def partition_fields_to_sql(self):
        """ """
        self.parse_partition_fields()
        if len(self.partition_fields) > 0:            
            partition_fields_content = str(self.partition_fields)[1:-1]
            self.sql_partition_fields = f"PARTITIONED BY ({partition_fields_content})"
        return self.sql_partition_fields

    def describe(self):        
        return TableUtils.describe_table(self.layer, self.table_name).display()

    def drop(self):
        return TableUtils.drop_table(self.layer, self.table_name)
    
    def exists(self):
        return TableUtils.table_exists(self.layer, self.table_name)
    
    def view_exists(self):
        return TableUtils.view_exists(self.layer, f"{self.table_name}_vw")
    
    def show_tbl_properties(self):
        if self.exists():
            return self.spark.sql(f"SHOW TBLPROPERTIES {self.layer}.{self.table_name}")
    
    def truncate(self):
        if self.exists():
            log(__name__).info(f"Truncating table {self.layer}.{self.table_name}.")
            return self.spark.sql(f"TRUNCATE TABLE {self.layer}.{self.table_name}")

    def parse(self):
        # generate template for create table
        self.parsed_table = """
            CREATE OR REPLACE TABLE {}.{} (
                {}
            )
            USING delta
            {}
            COMMENT "{}"
            {}
            LOCATION '{}'
        """.format(self.layer, self.table_name, self.fields_metadata_to_sql(), self.tbl_properties_to_sql(), self.table_comment_to_sql() , self.partition_fields_to_sql(), self.table_path)
        return self.parsed_table    
    
    def parse_view(self):        
        self.parse_fields_metadata()
        list_of_fields = []
        for field in self.fields_metadata:
            list_of_fields.append(field.hash_sql() if field.field_name in self.anonimized_fields else field.field_name)
        # generate template for create view
        self.parsed_view = """
            CREATE OR REPLACE VIEW {}.{}_vw (
                {}
            ) 
            COMMENT "{}"
            AS
            SELECT {} FROM {}.{}
        """.format(self.layer, self.table_name, self.fields_metadata_to_sql(), self.table_comment_to_sql(), ", ".join(list_of_fields), self.layer, self.table_name)
        return self.parsed_view
    
    def create(self):
        if not self.exists() or self.replace:                 
            ## generate final create table template            
            self.spark.sql(self.parse())                                            
            # mark table has been created or replaced by class
            self.has_created_table = True    
        else:
            log(__name__).warning(f"The table already exists, so nothing will be done.")
        return self
    
    def create_view(self):
        if not self.view_exists() or self.replace or self.has_created_table:
            ## generate final create view template            
            self.spark.sql(self.parse_view())                                            
            # mark table has been created or replaced by class            
        else:
            log(__name__).warning(f"The view already exists, so nothing will be done.")
        return self

    # alter methods
    def add_tbl_properties(self, tbl_properties):
        # armazenando o tbl_properties caso ele exista
        current_tbl_properties = self.tbl_properties 
        # setando temporariamente para usar o método interno de conversão
        self.tbl_properties = tbl_properties
        # rodando o alter table
        self.spark.sql("""
            ALTER TABLE {}.{} SET {}
        """.format(self.layer, self.table_name, self.tbl_properties_to_sql()))        
        # retornando as propriedades anteriores (se houvessem)
        self.tbl_properties = current_tbl_properties

    # remove tbl properties
    def remove_tbl_properties(self, list_of_tbl_properties_keys):        
        # rodando o alter table
        self.spark.sql("""
            ALTER TABLE {}.{} UNSET TBLPROPERTIES({})
        """.format(self.layer, self.table_name, str(list_of_tbl_properties_keys)[1:-1]))                
    
class BronzeTable(Table):

    layer = "bronze"

    def __init__(self,                  
                 table_name,
                 schema=None, 
                 yml={},    
                 partition_fields=[],       
                 anonimized_fields=[],  
                 tbl_properties={},                                                                      
                 comment={}, 
                 comment_squad={},  
                 fields={}):
        Table.__init__(self, self.layer, table_name, schema, yml, partition_fields, anonimized_fields, tbl_properties, comment, comment_squad, fields)          

class SilverTable(Table):

    layer = "silver"

    def __init__(self,                  
                 table_name, 
                 schema=None,
                 yml={}, 
                 partition_fields=[],   
                 anonimized_fields=[], 
                 tbl_properties={},
                 comment={}, 
                 comment_squad={},                                          
                 fields={}):
        Table.__init__(self, self.layer, table_name, schema, yml, partition_fields, anonimized_fields, tbl_properties, comment, comment_squad, fields)          

class GoldTable(Table,
                BatchTableWriter):

    layer = "gold"

    def __init__(self,                  
                 table_name, 
                 schema=None,
                 yml={}, 
                 partition_fields=[],    
                 anonimized_fields=[],                 
                 tbl_properties={},
                 comment={}, 
                 comment_squad={},          
                 fields={}):
        Table.__init__(self, self.layer, table_name, schema, yml, partition_fields, anonimized_fields, tbl_properties, comment, comment_squad, fields)          
        BatchTableWriter.__init__(self, self.layer, table_name)
            



from great_expectations.dataset import SparkDFDataset
from great_expectations.dataset import MetaSparkDFDataset
from great_expectations.core.expectation_suite import ExpectationSuite 
from great_expectations.core.expectation_configuration import ExpectationConfiguration
from dnalib.utils import Utils
from pyspark.sql.functions import col

# Criando uma classe custom de Dataset, é uma implementação com maior 
# liberdade para adicionar expectations novos.
class GreatExpectations(SparkDFDataset):    

    """ 
        Exemplo de classe com expectations customizados para dataframes SQL do PySpark.
    """    
    default_suite_name = "ge_custom_validator"
    def __init__(self, df_or_query, **kwargs):           
        if isinstance(df_or_query, str):
            spark_df = Utils.spark_instance().sql(df_or_query)
        else:
            spark_df = df_or_query  
        SparkDFDataset.__init__(self, spark_df=spark_df, **kwargs)
        self.ge_suite = ExpectationSuite(self.default_suite_name)   
        self.ge_result = None     

    # Método que valida se os valores de uma coluna estão dentro de um conjunto (lista),
    # caso não esteja retorna em partial_unexpected_list (amostras das colunas definidas em keys_to_return)
    @MetaSparkDFDataset.expectation(["column", "value_set", "keys_to_return", "limit_return", "mostly"])
    def expect_column_values_to_be_in_set_given_keys_to_return(self, column, value_set, keys_to_return=[], limit_return=100, mostly=0.99, **kwargs):
        df_verify = self.spark_df.withColumn("__success", col(column).isin(value_set))
        partial_unexpected_list = df_verify.filter(col("__success") == False)      
        number_of_wrong = partial_unexpected_list.count()  
        rows_to_verify = df_verify.count()
        return {
            "success" : (number_of_wrong / rows_to_verify) >= mostly,
            "result" : {
                "partial_unexpected_list": partial_unexpected_list.select(keys_to_return).head(limit_return),
                "unexpected_count": number_of_wrong
            }
        }


    # Método que valida se o campo é um cpf ou cnpj válido (considerando apenas números,cpf de 11 e 6 digitos (miolo do cpf) e cnpj de 14 digitos)
    @MetaSparkDFDataset.expectation(["column", "mostly"])
    def expect_column_to_be_cpf_or_cnpj_number(self, column, mostly=0.99):
        return super().expect_column_values_to_match_regex_list(
            column=column,
            regex_list=['(^[0-9]{14}$)|(^[0-9]{11}$)|(^[0-9]{6}$)'],
            match_on='any',
            mostly=mostly 
        )
    
    # Método que valida se as colunas do dataframe seguem certos padrões de prefixos
    @MetaSparkDFDataset.expectation(["columns", "columns_prefix_patterns"])
    def expect_columns_to_follow_prefix_patterns(self, columns, columns_prefix_patterns):                             
        columns_wrong_prefix = []
        for column_name in columns:
            if not column_name.startswith(tuple(columns_prefix_patterns)):
                columns_wrong_prefix.append(column_name)
        return {
            "success": len(columns_wrong_prefix) == 0,
            "result": {    
                "partial_unexpected_list": columns_wrong_prefix,            
                "unexpected_count": len(columns_wrong_prefix)
            }
        }

    # Método que valida se as colunas do dataframe seguem padrões de nomenclatura
    @MetaSparkDFDataset.expectation(["columns", "columns_valid_types"])
    def expect_columns_to_be_of_valid_types(self, columns, columns_valid_types):       
        invalid_type_fields = []
        for field in self.spark_df.schema:
            field_name = field.name
            field_type = field.dataType.simpleString()    
            if not field_type in columns_valid_types and field_name in columns:
                invalid_type_fields.append(f"Invalid type {field_type} for field {field_name}.") 
        return {
            "success": len(invalid_type_fields) == 0,
            "result": {
                "partial_unexpected_list": invalid_type_fields,
                "unexpected_count": len(invalid_type_fields)
            }
        }

    # Método que valida se as colunas do dataframe seguem tipos validos
    @MetaSparkDFDataset.expectation(["columns"])
    def expect_columns_to_be_commented(self, columns):  
        empty_comment_columns = []
        for field in self.spark_df.schema:
            field_name = field.name
            field_comment = field.metadata.get("comment", "").strip()                    
            if field_comment == "" and field_name in columns:
                empty_comment_columns.append(f"The field {field.name} has no comment")
        return {
            "success": len(empty_comment_columns) == 0,
            "result": {
                "partial_unexpected_list": empty_comment_columns,
                "unexpected_count": len(empty_comment_columns)
            }
        }

    # Método que valida se as colunas do dataframe seguem tipos validos
    @MetaSparkDFDataset.expectation(["columns"])
    def expect_columns_to_be_named_in_lower_case(self, columns):  
        columns_not_in_lower_case = []
        for column_name in columns:
            if not column_name.islower():
                columns_not_in_lower_case.append(column_name)
        return {
            "success": len(columns_not_in_lower_case) == 0,
            "result": {
                "partial_unexpected_list": columns_not_in_lower_case,
                "unexpected_count": len(columns_not_in_lower_case)
            }
        }

    def add_expectation(self, expectation_name, expectations_kwargs):        
        """ """
        self.ge_suite.append_expectation(ExpectationConfiguration(expectation_name, expectations_kwargs))

    def validate(self):
        """ """
        self.ge_result = super().validate(self.ge_suite)
        return self.ge_result



from pyspark.sql.functions import col, to_json
from dnalib.utils import Utils
from .loader import Loader
from dnalib.log import *
from firebase_admin import credentials, firestore, initialize_app
from google.api_core.retry import Retry
from tempfile import NamedTemporaryFile
import json

class FireBaseApp: 
    """
        A classe FireBaseApp implementa um singleton para manter uma instancia única de execução do cliente do Firestore.
    """
    _instance = None

    def __init__(self, service_account_data, db_name=None):
        self.service_account_data = service_account_data
        self.db_name = db_name
        self.db = self.authenticate_db()
        
    def authenticate_db(self):
        with NamedTemporaryFile('w+') as fp:
            fp.write(self.service_account_data)
            fp.seek(0)
            cred = credentials.Certificate(fp.name)
            initialize_app(cred)
            return firestore.client(database_id=self.db_name)
        return None
    
    @classmethod
    def instance(cls, service_account_data, db_name=None):
        if cls._instance is None:
            cls._instance = cls(service_account_data, db_name)
        return cls._instance

class FireBaseLoader(Loader):
    """
       A classe FireBaseLoader encapsula o processo de carregamento de dados do Firestore para um dataframe Spark. 
    """
    def __init__(self, 
                 source_path,       
                 loader_type,           
                 collection_name, 
                 order_by='__name__', 
                 page_size=1000,
                 init_load=None,
                 end_load=None,
                 flat_json=True,
                 db_name=None,
                 spark_num_partitions=None):
        super().__init__(source_path, loader_type)
        self.collection_name = collection_name
        self.order_by = order_by
        self.page_size = page_size
        self.init_load = init_load
        self.end_load = end_load
        self.flat_json = flat_json
        self.db_name = db_name
        self.spark_num_partitions = spark_num_partitions
        self.service_account_data = Utils.key_vault_value(source_path)   
        self.db = FireBaseApp.instance(self.service_account_data, self.db_name).db
        self.collection = self.load_collection()
        self.docs = []
        self.count = 0
        self.df_raw = None            

    def load_collection(self):
        """  """
        collection = self.db.collection(self.collection_name)
        if self.init_load:
            filter_init = firestore.FieldFilter(self.order_by, ">=", self.init_load)
            collection = collection.where(filter=filter_init)
        if self.end_load:
            filter_end = firestore.FieldFilter(self.order_by, "<=", self.end_load)
            collection = collection.where(filter=filter_end)
        return collection

    def iterate_documents(self):
        """  """
        cursor = None
        self.count = 0
        while True:
            docs = []  
            if cursor:
                docs = [snapshot for snapshot in self.collection.limit(self.page_size).order_by(self.order_by).start_after(cursor).stream(retry=Retry())]
            else:
                docs = [snapshot for snapshot in self.collection.limit(self.page_size).order_by(self.order_by).stream(retry=Retry())]
            self.count = self.count + len(docs)
            for doc in docs:
                # return doc to iterate  
                yield doc                
            print(f"Firebase loaded {self.count} documents.", end="\r")            
            if len(docs) == self.page_size:
                cursor = docs[self.page_size-1]
                continue
            break
    
    def load_incremental(self):
        """  """       
        for doc in self.iterate_documents():            
            self.docs.append(json.dumps(doc.to_dict(), default=str))
        log(__name__).info(f"Firebase loaded {self.count} documents.")
        # converts all documents to json string, make primitives string to avoid break schema
        if self.flat_json:
            par = Utils.spark_instance().sparkContext.parallelize(self.docs, self.spark_num_partitions)
            self.df_raw = Utils.spark_instance().read.json(par, primitivesAsString=True) 
            for field in self.df_raw.schema:
                if field.dataType.simpleString() != "string":
                    self.df_raw = self.df_raw.withColumn(field.name, to_json(col(field.name)))   
        else:            
            par = Utils.spark_instance().sparkContext.parallelize([[doc] for doc in self.docs], self.spark_num_partitions)
            self.df_raw = Utils.spark_instance().createDataFrame(par, "json_document: string")             
        return self.df_raw
    
    def load_full(self):
        """  """
        return self.load_incremental()
    